//AlicePF.java - Alice's part of the 2-party PF-SFE protocol. 
//Copyright (c) 2007 Thomas Schneider
//See full Copyright license terms in file ../GPL.txt

package SFE.PF;

import java.io.*;
import java.net.*;
import org.apache.log4j.*;

import SFE.BOAL.Circuit;
import SFE.BOAL.Gate;
import SFE.BOAL.IO;
import SFE.BOAL.OT;
import SFE.BOAL.MyUtil;
import SFE.BOAL.Formatter;
import SFE.BOAL.FormatterError;


/**
 * Alice's part of the 2-party PF-SFE protocol.
 *
 * @author Thomas Schneider
 */
public class AlicePF {
	private static final Logger logger = Logger.getLogger(AlicePF.class);
	Formatter2 f = null;
	int nInputs, nOutputs, nGates;
	Gate[] UC_outputs;
	
	class Formatter2 extends Formatter {
		private static final long serialVersionUID = 7372280076169056025L;

		int nAliceInputs=0, nAliceOutputs=0;
		int nBobInputs=0, nBobOutputs=0;
		
		public void getAliceOutput(Circuit c) {
			int bit;
			int value;
			IO io = null;

			// Scan through all format objects 
			for (int i = 0; i < FMT.size(); i++) {
				io = (IO) FMT.elementAt(i);

				// Verify it's an output object of the approriate player (i.e., Alice/Bob).
				if (io.isInput() || !io.isAlice()) {
					continue;
				}

				// collect all specified output bits, one by one
				for (int j = value = 0; j < io.getNLines(); j++) {
					// this is the value of the output gate at line_num
					int line_num=io.getLinenum(j);
					bit = UC_outputs[line_num].getValue();
					value |= (bit << j);
				}
				
				System.out.println(io.getPrefix() + value);
			}
		}		
		
		public byte[] fextractOutPayload (Circuit c, int total_size, boolean alice_inputs) {
			int k=0;
			byte small_byte_arr[];
			byte big_byte_arr[]=new byte[total_size];
			
			if(alice_inputs) {
				logger.error("This method is not implemented.");
				System.exit(1);
			}

			// Extract payload data from relevant gates,
			// and store them in the result array
			for(int i=nOutputs-1; i>=nAliceOutputs; i--) {
				Gate g = UC_outputs[i];
				small_byte_arr = g.gextractOutPayload();
				for (int j1 = 0; j1 < small_byte_arr.length ; j1++, k++)
					big_byte_arr[k] = small_byte_arr[j1];
			}

			return (big_byte_arr);
		}
		
	    public void finjectInpPayload (Circuit c, byte[] info, boolean alice_inputs) {
	    	int k=0;
	    	if(alice_inputs) {
				logger.error("This method is not implemented.");
	    		System.exit(1);
	    	}
	    	for (int i=nAliceInputs+nBobInputs-1; i>=nAliceInputs; i--) {
                Gate g = c.getGate(i);
                k += g.ginjectInpPayload(info, k);
	    	}
	    }
		
	    /**
	     * This routine marks all the input/output gates of a circuit as
	     * belonging to either Alice or Bob.
	     *
	     * @param c a gate-level circuit
	     */
	    public void markIO (Circuit c, int[] tsize) {

	        tsize[0] = 0;
	        tsize[1] = 0;
	        
	        // Alice's inputs
	        for(int i=0; i<nAliceInputs; i++) {
	        	Gate g = c.getGate(i);
	        	g.markAliceBob(true);
	        }
	        
	        // Bob's inputs
	        for(int i=nInputs-1; i>=nAliceInputs; i--) {
	        	Gate g = c.getGate(i);
	        	g.markAliceBob(false);
	        	tsize[0] += g.gmeasureInpPayload();
	        }
	        
	        // Alice's outputs
	        for(int i=0; i<nAliceOutputs; i++) {
	        	Gate g = UC_outputs[i];
	        	g.markAliceBob(true);
	        }
	        
	        // Bob's outputs
	        for(int i=nOutputs-1; i>=nAliceOutputs; i--) {
	        	Gate g = UC_outputs[i];
	        	g.markAliceBob(false);
	        	tsize[1] += g.gmeasureOutPayload();
	        }
	    }
		
		protected void parseNumbers() throws FormatterError {
			int val = parseToken(false);
			
			if (val == WORD_INDICATOR) {
                throw new FormatterError("parseNumbers : bad word " + s);
	        } else {
	        	// Store this line-number in the current IO object
	        	for(int i=0; i<val; i++) {
	        		if(curIO.isInput()) curIO.addLinenum(nAliceInputs++);
	        		else curIO.addLinenum(nAliceOutputs++);
	        	}
	        }
		}
		
		private void parseUC() throws FormatterError {
			int val;
			
			parseString("inputs");
			val = parseToken(false);
			if (val == WORD_INDICATOR) 
				throw new FormatterError("parseUC : bad word " + s);
			nInputs=val;
			
			parseString("outputs");	
			val = parseToken(false);
			if (val == WORD_INDICATOR)
				throw new FormatterError("parseUC : bad word " + s);
			nOutputs=val;

			parseString("gates");			
			val = parseToken(false);
			if (val == WORD_INDICATOR) 
				throw new FormatterError("parseUC : bad word " + s);
			nGates=val;
		}
		
	    private boolean parseLine() throws FormatterError {
	        int ln = parseToken(true);

	        if (ln == EOF_INDICATOR) {
	            return false;
	        }

	        if (ln != WORD_INDICATOR) {
	            throw new FormatterError("parseAliceBob : Bad first word");
	        }
	        
	        if(s.equals("Alice")) {
		        // allocate new IO object for Alice
		        curIO = new IO(true);
		        parseLineName();
	        }
	        else if(s.equals("UC")) {
	        	parseUC();
	        }
	        else {
	        	throw new FormatterError("parseFirst : Bad first word");
	        }

	        return (true);
	    }
		
	    public void parse() throws FormatterError {
	        try {
	            while (parseLine());
	            logger.info("Parse done.");
	            
	            nBobInputs=nInputs-nAliceInputs;
	            nBobOutputs=nOutputs-nAliceOutputs;
	            
	            logger.debug("nAliceInputs="+nAliceInputs+", nBobInputs="+nBobInputs+", nAliceOutputs="+nAliceOutputs+", nBobOutputs="+nBobOutputs);
	            
	            st = null;
	        } catch (Exception e) {
	            logger.error("parse: exception " + e.getMessage());
	            e.printStackTrace();
	        }
	    }
		
		Formatter2(StreamTokenizer s) {
			super(s);
		}
	}
	
	//---------------------------------------------------------------

	/**
	 * AlicePF Constructor
	 *
	 * @param fmtAlice_filename - format filename
	 * @param hostname - where to find Bob
	 * @param num_iterations - how many iterations to do
	 * @param stats - print run statistics in the end
	 */
	public AlicePF(String fmtAlice_filename, String sseed,
			String hostname, int num_iterations, boolean stats, String UC_type) throws Exception {
		UC c=null;
		BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
		int i;
		int ot_type;
		int[] bob_io_size = new int[2];
		OT ot;
		Socket sock = null;
		ObjectInputStream fromBob = null;
		ObjectOutputStream toBob = null;
		long sum1=0, sum2=0, sum3=0, sum4=0;
		long sum1a=0, sum1b=0, sum1c=0, sum1d=0;

		// Preparations
		MyUtil.init(sseed);

		// Connect to Bob
		try {
			sock = new Socket(hostname, 3496);
			fromBob = new ObjectInputStream(sock.getInputStream());
			toBob = new ObjectOutputStream(sock.getOutputStream());
		} catch (UnknownHostException e) {
			System.err.println("AlicePF: Don't know host " + hostname);
			System.exit(-1);
		} catch (IOException e) {
			System.err.println(
					"AlicePF: Couldn't get I/O for the connection with Bob");
			System.exit(-1);
		}

		MyUtil.sendInt(toBob, num_iterations, true);
		ot_type = MyUtil.receiveInt(fromBob);
		ot = new OT(ot_type);

		byte[] EncPayload;
		byte[] InpPayload;
		byte[] OutPayload;
		int InpPayloadSize=0;
		int OutPayloadSize=0;
		int EncPayloadSize=0;

		for (i = 0; i < num_iterations; i++) {

			MyUtil.deltaTime (true);

			logger.info("----------------------------------------------");
			logger.info("Iteration no = " + i);

			// Parse the IOformat file and prepare the inputs
			try {
				// Preparations
				FileReader fmt = new FileReader(fmtAlice_filename);
				StreamTokenizer fmtst = new StreamTokenizer(fmt);

				// IO Formatting
				f = new Formatter2(fmtst);
				f.parse();

				// Cleanup
				fmt.close();
			} catch (IOException e) {
				logger.error("AlicePF: cannot open/close " + fmtAlice_filename + " - " +
						e.getMessage());
			} catch (FormatterError e) {
				logger.error("AlicePF: parsing " + fmtAlice_filename + " failed.");
			} catch (Exception e) {
				logger.error("AlicePF: exception - " + e.getMessage());
			}

			sum1a += MyUtil.deltaTime (false);
			
			// generate UC
			c=UC.generateUC(nInputs, nOutputs, nGates, UC_type);
			sum1b += MyUtil.deltaTime (false);
			
			UC_outputs=c.extractCircuit(null);
			sum1c += MyUtil.deltaTime (false);

			// Mark its inputs & outputs
			f.markIO (c, bob_io_size); 
			InpPayloadSize = bob_io_size[0];
			OutPayloadSize = bob_io_size[1];
			logger.debug("InpPayloadSize="+InpPayloadSize+" OutPayloadSize="+OutPayloadSize);


			// Encrypt it (dummy) to determine size
			c.generateEncCircuit();
			EncPayloadSize = c.cmeasureEncPayload();

			sum1d += MyUtil.deltaTime (false);
			
			// Run the SFE protocol
			// ====================

			// Receive encrypted Circuit from Bob
			EncPayload = new byte[EncPayloadSize];
			MyUtil.receiveBytes (fromBob, EncPayload, EncPayloadSize);			
			c.cinjectEncPayload(EncPayload);
			
			// Receive Bob's inputs and inject them in UC
			InpPayload = new byte[InpPayloadSize];
			MyUtil.receiveBytes (fromBob, InpPayload, InpPayloadSize);
			f.finjectInpPayload (c, InpPayload, false);

			sum2 += MyUtil.deltaTime (false);

			// Read Alice's inputs
			f.getAliceInput(c, br);

			MyUtil.deltaTime (true);
			// OTs - Alice is the chooser + 
			// place Alice's inputs in the chosen circuit
			logger.info("starting OT...");
			OT.ChooserOTs(c, f, ot, toBob, fromBob);
			logger.info("OT completed.");

			sum3 += MyUtil.deltaTime (false);

			// Evaluate UC
			logger.info("evaluating garbled circuit...");
			c.evalGarbledCircuit(true, false);
			logger.info("circuit evaluation completed.");

			// Send Bob his garbled results
			OutPayload = f.fextractOutPayload (c, OutPayloadSize, false);
			MyUtil.sendBytes (toBob, OutPayload, true);

			// print Alice's output
			f.getAliceOutput(c);

			sum4 += MyUtil.deltaTime (false);
		}

		if (stats) {
			sum1=sum1a+sum1b+sum1c+sum1d;
			System.out.println("Initial calculations   [sum1]  = " + (float)sum1/1000.0);
			System.out.println(" Generate UC            [sum1b] = " + (float)sum1b/1000.0);
			System.out.println(" Extract Circuit        [sum1c] = " + (float)sum1c/1000.0);
			System.out.println(" Get Encrypted Size     [sum1d] = " + (float)sum1d/1000.0);
			System.out.println("Circuits communication [sum2]  = " + (float)sum2/1000.0);
			System.out.println("Oblivious Transfers    [sum3]  = " + (float)sum3/1000.0);
			System.out.println("Evaluation & output    [sum4]  = " + (float)sum4/1000.0);
		}

		// Cleanup
		toBob.close();
		fromBob.close();
		sock.close();
	}

	//---------------------------------------------------------------

	/**
	 * This routine is for debugging socket communication
	 */
	public void pongping(ObjectOutputStream toBob, ObjectInputStream fromBob,
			int a) {
		System.out.println("Sending " + a + " to Bob");
		MyUtil.sendInt(toBob, a, true);
		System.out.println("Attempting to read num from Bob");

		int u = MyUtil.receiveInt(fromBob);
		System.out.println("Got Int from Bob " + u);
	}

	//---------------------------------------------------------------
	/**
	 * Display usage of AlicePF
	 * <p>
	 * Usage: java SFE.PF.AlicePF -r|-s &lt;filename&gt; &lt;seed&gt; &lt;hostname&gt; [num_iterations] [UC_type]<br>
	 * -r = RUN, -s = RUNSTATS
	 * <p>
	 * Examples:
	 * <ol>
	 *   <li>java SFE.PF.AlicePF -r Maximum.txt Xb@&5H1m!p localhost</li>
	 *   <li>java SFE.PF.AlicePF -s Maximum.txt Xb@&5H1m!p localhost 10 UCpract</li>
	 * </ol>
	 */
	public static void aliceUsage(int err_code) {
		//System.out.println("AlicePF activation error code = " + err_code);
		System.out.println("Usage: java SFE.PF.AlicePF -r|-s <filename> <seed> <hostname> [num_iterations] [UC_type]");
		System.out.println(" -r = RUN, -s = RUNSTATS");
		System.out.println(" Examples: 1. java SFE.PF.AlicePF -r Maximum.txt Xb@&5H1m!p localhost");
		System.out.println("           2. java SFE.PF.AlicePF -s Maximum.txt Xb@&5H1m!p localhost 10 UCpract");
		System.exit(1);
	}

	//---------------------------------------------------------------

	/**
	 * Main program for activating AlicePF
	 *
	 * @param args - command line arguments.<br>
	 *               args[0] should be -r or -s<br>
	 *               args[1] should be filename<br>
	 *               args[2] should be seed for RNG<br>
	 *               args[3] should be hostname<br>
	 *               args[4] can be number of iterations<br>
	 *               args[5] can be type of UC
	 *
	 * @see SFE.PF.AlicePF#aliceUsage
	 */
	public static void main(String[] args) throws Exception {
		String filename;
		String fmtAlice_fname;
		int num_iterations;
		boolean run_stats = false;
		boolean run = false;

		// Load logging configuration file
		PropertyConfigurator.configure(MyUtil.pathFile("SFE_logcfg.lcf"));

		// Various legality tests on command line parameters

		if (args.length < 4)
			aliceUsage(1);

		run_stats = args[0].equals("-s") ;
		run = run_stats || args[0].equals("-r") || args[0].equals("-rn");

		if (!run && !run_stats)
			aliceUsage(2);

		filename = new String(args[1]);
		fmtAlice_fname = new String(filename + ".shdlPF.Alice.Opt.fmt");

		if (run) {
			File f1 = new File(fmtAlice_fname);

			if (!f1.exists()) {
				System.out.println("Input file " + fmtAlice_fname + " not found");
				aliceUsage(3);
			}
		}

		// Do something (finally...)

		if (run) {
			System.out.println("Running AlicePF...");
			String UC_type;
			try {
				if (args.length >= 5)
					num_iterations = Integer.parseInt(args[4]);
				else
					num_iterations = 1 ;

				if(args.length == 6)
					UC_type=args[5];
				else
					UC_type=null;
				
				new AlicePF(fmtAlice_fname, args[2], args[3], num_iterations, run_stats, UC_type);
			} catch (Exception e) {
				System.out.println("AlicePF's main err: " + e.getMessage());
				e.printStackTrace();
			}
		}
	}
}
