//BobPF.java - Bob's part of the 2-party PF-SFE protocol.
//Copyright (c) 2007 Thomas Schneider
//See full Copyright license terms in file ../GPL.txt

package SFE.PF;

import SFE.Compiler.*;
import SFE.PF.ParseGate;

import java.io.*;
import java.net.*;
import java.util.*;
import org.apache.log4j.*;
import SFE.BOAL.Gate;
import SFE.BOAL.IO;
import SFE.BOAL.Parser;
import SFE.BOAL.OT;
import SFE.BOAL.MyUtil;
import SFE.BOAL.Formatter;
import SFE.BOAL.FormatterError;

/**
 * Bob's part of the 2-party PF-SFE protocol.
 *
 * @author Thomas Schneider
 */
public class BobPF {
	private static final Logger logger = Logger.getLogger(BobPF.class);
	Formatter2 f;
	
	class Formatter2 extends Formatter {
		private static final long serialVersionUID = 2302928117362964316L;

	    /**
	     * This routine marks all the input/output gates of a circuit as
	     * belonging to either Alice or Bob.
	     *
	     * @param c a gate-level circuit
	     */
	    public void markIO(UC c, int[] tsize) {

	        tsize[0] = 0;
	        tsize[1] = 0;

	        // scan through all format object for input objects
	        for (int i = 0; i < FMT.size(); i++) {

	            IO io = (IO) FMT.elementAt(i);

	            // mark all input/output wires
	            for (int j = 0; j < io.getNLines(); j++) {
	                int line_num = io.getLinenum(j);
	                int translated_line_num = c.translateIO(line_num, io.isAlice(), io.isInput());
	                io.setLinenum(j,translated_line_num);
	                Gate g = c.getGate(translated_line_num);
	                g.markAliceBob (io.isAlice());
	                if (g.isBobInput())  tsize[0] += g.gmeasureInpPayload();
	                if (g.isBobOutput()) tsize[1] += g.gmeasureOutPayload();
	            }
	        }
	    }

	    private boolean parseLine() throws FormatterError {
	        int ln = parseToken(true);

	        if (ln == EOF_INDICATOR) {
	            return false;
	        }

	        if (ln != WORD_INDICATOR) {
	            throw new FormatterError("parseAliceBob : Bad first word");
	        }
	        
	        if(s.equals("Alice")) {
		        curIO = new IO(true);
		        parseLineName();
	        }
	        else if(s.equals("Bob")){
		        curIO = new IO(false);
		        parseLineName();        	
	        }
	        else {
	        	throw new FormatterError("parseFirst : Bad first word");
	        }

	        return (true);
	    }
		
	    public void parse() throws FormatterError {
	        try {
	            while (parseLine());
	            logger.info("Parse done.");

	            st = null;
	        } catch (Exception e) {
	            logger.error("parse: exception " + e.getMessage());
	            e.printStackTrace();
	        }
	    }
		
		Formatter2(StreamTokenizer s) {
			super(s);
		}
	}	

	class Parser2 extends Parser {
		Vector ins, outs, gates;
		
		private int[] parseInputs() throws ParseError {
	    	int x[]=new int[2];
	    	
	        for (int i = 0; i < 2; i++) {
	            x[i]=parseWireNum();
	        }

	        return x;
	    }
		
	    private int[] parseTruthTable() throws ParseError {
	        int out_bit;

	        int x[]=new int[4];
	    	        
	        // expecting 4 output bits (0/1)
	        for (int i = 0; i < 4; i++) {
	            out_bit = parseToken(false);
	            
	            if ((out_bit != 0) && (out_bit != 1)) {
	                throw new ParseError("ParseOutputs: bad output bit at " +
	                    line_num);
	            }
	            
	            x[i]=out_bit;
	        }

	        return x;
	    }

		private void parseGate2(boolean out_line) throws ParseError {
	        int[] truthTable, inputs;
	        
	        parseString("arity");
	        ninputs = parseNumInputs();
	        if(ninputs!=2)
	    		throw new ParseError("parseGate2 : SHDL2 circuits can only contain 2-input gates at line " +
	                    line_num);

	        parseString("table");
	        truthTable=parseTruthTable();
	        parseString("inputs");
	        inputs=parseInputs();
	        
	        gates.add(new ParseGate(truthTable,inputs,out_line));
	    }
				
	    public void parse() throws ParseError {
	        boolean out_line;
	        int line_type;

	        try {
	            while (parseLineNum()) {
	                out_line = parseOptString("output");
	                if(out_line) {
	                	outs.add(new Integer(line_num));
	                }
	                line_type = parseLineName();

	                if (line_type == INPUT_LINE) {
	                	ins.add(new Integer(line_num));
	                } else {
	                    parseGate2(out_line);
	                }
	            }
	        } catch (Exception e) {
	            logger.error("parse: exception " + e.getMessage());
	        }
	    }
	    
	    public Parser2(StreamTokenizer st, Vector ins, Vector outs, Vector gates) {
			super(st);
			this.ins=ins;
			this.outs=outs;
			this.gates=gates;
		}
	}
	
	//---------------------------------------------------------------   

	/**
	 * BobPF Constructor
	 *
	 * @param circuit_filename - circuit filename
	 * @param fmt_filename - format filename
	 * @param sot_type - type of OT to perform (String)
	 * @param stats - print run statistics in the end
	 */
	public BobPF(String circuit_filename, String fmt_filename, String sseed, String sot_type, boolean stats, String UC_type)
	throws IOException {
		UC c=null;
		BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
		int i;
		int num_of_iterations;
		int ot_type = Integer.parseInt(sot_type);
		int[] bob_io_size = new int[2];
		Parser p = null; // the parser and builder of the circuit
		OT ot;
		ServerSocket SS = null;
		Socket sock = null;
		ObjectInputStream fromAlice = null;
		ObjectOutputStream toAlice = null;
		long sum1=0, sum2=0, sum3=0;
		long sum1a=0, sum1b=0, sum1c=0, sum1d=0, sum1e=0, sum1f=0;

		// Preparations
		MyUtil.init(sseed);

		// Establish connection with Alice
		try {
			SS = new ServerSocket(3496); // create a server socket
			logger.info("BobPF: waiting for Alice to connect");
			sock = SS.accept(); // Accept Alice on this socket
			toAlice = new ObjectOutputStream(sock.getOutputStream());
			fromAlice = new ObjectInputStream(sock.getInputStream());
		} catch (IOException e) {
			System.out.println(
					"BobPF: establishing connection with Alice failed: " +
					e.getMessage());
			System.exit(-1);
		}

		MyUtil.sendInt(toAlice, ot_type, true);
		ot = new OT(ot_type);
		int aint = MyUtil.receiveInt(fromAlice);
		num_of_iterations = aint;

		byte[] EncPayload;
		byte[] InpPayload;
		byte[] OutPayload;
		int EncPayloadSize=0;
		int InpPayloadSize=0;
		int OutPayloadSize=0;

		for (i = 0; i < num_of_iterations; i++) {

			MyUtil.deltaTime (true);
			
			logger.info("----------------------------------------------");
			logger.info("Iteration no = " + i);

			// Parse the IOformat file and prepare the inputs
			try {
				// Preparations
				FileReader fmt = new FileReader(fmt_filename);
				StreamTokenizer fmtst = new StreamTokenizer(fmt);

				// IO Formatting
				f = new Formatter2(fmtst);
				f.parse();

				// Cleanup
				fmt.close();
			} catch (IOException e) {
				logger.error("Bob: cannot open/close " + fmt_filename + " - " +
						e.getMessage());
			} catch (FormatterError e) {
				logger.error("Bob: parsing " + fmt_filename + " failed.");
			} catch (Exception e) {
				logger.error("Bob: exception - " + e.getMessage());
			}

			FileReader fr = null;
			StreamTokenizer st = null;

			logger.info("Reading SHDL2 circuit file.");
			
			// Preparations
			try {
				fr = new FileReader(circuit_filename);
				st = new StreamTokenizer(fr);
			} catch (IOException e) {
				logger.error("main: cannot open " + circuit_filename + " - " +
						e.getMessage());
			}
			
			int ins[]=null;
			int outs[]=null;
			ParseGate gates[]=null;
			try {
				Vector vins=new Vector(1000,1000);
				Vector vouts=new Vector(1000,1000);
				Vector vgates=new Vector(1000,1000);

				// Parse
				p = new Parser2(st,vins,vouts,vgates);
				p.parse();

				// Clean-up
				fr.close();
				
				ins=new int[vins.size()];
				for(int j=0; j<vins.size(); j++)
					ins[j]=((Integer)vins.get(j)).intValue();
				
				outs=new int[vouts.size()];
				for(int j=0; j<vouts.size(); j++)
					outs[j]=((Integer)vouts.get(j)).intValue();
				
				gates=new ParseGate[vgates.size()];
				for(int j=0; j<vgates.size(); j++)
					gates[j]=(ParseGate)vgates.get(j);
				
			} catch (Exception e) {
				logger.error("main: exception - " + e.getMessage());
				System.exit(1);
			}			

			sum1a += MyUtil.deltaTime (false);
			
			// Generate UC
			c=UC.generateUC(ins.length, outs.length, gates.length, UC_type);
			sum1b += MyUtil.deltaTime (false);			
			c.extractCircuit(null);
			sum1c += MyUtil.deltaTime (false);

			// mark inputs & outputs
			f.markIO(c, bob_io_size);
			InpPayloadSize = bob_io_size[0];
			OutPayloadSize = bob_io_size[1];
			logger.info("InpPayloadSize="+InpPayloadSize+" OutPayloadSize="+OutPayloadSize);

			sum1d += MyUtil.deltaTime (false);
			// Program UC
			c.program(ins, outs, gates);
			sum1e += MyUtil.deltaTime (false);
			ins=outs=null;
			gates=null;
			
			// Run the SFE protocol
			// ====================
			
			// Encrypt UC
			c.generateEncCircuit();
			sum1f += MyUtil.deltaTime (false);

			// Send encrypted UC to Alice
			EncPayloadSize = c.cmeasureEncPayload();
			logger.debug("EncPayloadSize="+EncPayloadSize);
			EncPayload = c.cextractEncPayload(EncPayloadSize);
			MyUtil.sendBytes(toAlice, EncPayload, false);
			toAlice.flush();
			sum2 += MyUtil.deltaTime (false);
			
			// Read Bob's inputs + update circuit accordingly
			f.getBobInput(c, br);

			// Send Bob's inputs (garbled) to Alice
			InpPayload = f.fextractInpPayload(c, InpPayloadSize, false);
			MyUtil.sendBytes(toAlice, InpPayload, true);
			sum2 += MyUtil.deltaTime (false);

			// OTs - Bob is the sender: send Alice's inputs
			OT.SenderOTs(c, f, ot, toAlice, fromAlice);
			sum3 += MyUtil.deltaTime (false);

			// Get Bob garbled results from Alice & print them
			OutPayload = new byte[OutPayloadSize];
			MyUtil.receiveBytes (fromAlice, OutPayload, OutPayloadSize);
			f.finjectOutPayload (c, OutPayload, false);
			f.getBobOutput(c);
		}

		if (stats) {
			sum1=sum1a+sum1b+sum1c+sum1d+sum1e+sum1f;
			System.out.println("Initial calculations   [sum1]  = " + (float)sum1/1000.0);
			System.out.println(" Generate UC            [sum1b] = " + (float)sum1b/1000.0);
			System.out.println(" Extract Circuit        [sum1c] = " + (float)sum1c/1000.0);
			System.out.println(" Program UC             [sum1e] = " + (float)sum1e/1000.0);
			System.out.println(" Encrypt Circuit        [sum1f] = " + (float)sum1f/1000.0);
			System.out.println("Circuits communication  [sum2]  = " + (float)sum2/1000.0);
			System.out.println("Oblivious transfer      [sum3]  = " + (float)sum3/1000.0);
		}

		// Cleanup
		toAlice.close();
		fromAlice.close();
		sock.close();
		SS.close();
	}

	//---------------------------------------------------------------

	/**
	 * This routine is for debugging socket communication
	 */
	public void pingpong(ObjectOutputStream toAlice,
			ObjectInputStream fromAlice, int a) {
		System.out.println("Attempting to read num from Alice");

		int u = MyUtil.receiveInt(fromAlice);
		System.out.println("Got Int from Alice " + u);
		System.out.println("Sending " + a + " to Alice");
		MyUtil.sendInt(toAlice, a, true);
	}

	//---------------------------------------------------------------
	/**
	 * Display usage of BobPF
	 * <p>
	 * Usage: java SFE.PF.BobPF -c|-r|-s &lt;filename&gt; &lt;seed&gt; <ot_type> [UC_type]<br>
	 * -c = COMPILE, -r = RUN, -s = RUNSTATS<br>
	 * (&lt;seed&gt;, &lt;ot_type&gt; expected only with -r)
	 * <p>
	 * Examples:
	 * <ol>
	 *   <li>java SFE.PF.BobPF -c Maximum.txt</li>
	 *   <li>java SFE.PF.BobPF -r Maximum.txt bQ91:d_aV!|l 4</li>
	 *   <li>java SFE.PF.BobPF -s Maximum.txt bQ91:d_aV!|l 4 UCpract</li>
	 * </ol>
	 */
	public static void bobUsage(int err_code) {
		//System.out.println("BobPF activation error code = " + err_code);
		System.out.println("Usage: java SFE.PF.BobPF -c|-r <filename> <seed> <ot_type> [UC_type]");
		System.out.println("-c = COMPILE, -r = RUN, -s = RUNSTATS");
		System.out.println("(<seed>, <ot_type> expected only with -r)") ;
		System.out.println("Examples: 1. java SFE.PF.BobPF -c Maximum.txt");
		System.out.println("          2. java SFE.PF.BobPF -r Maximum.txt bQ91:d_aV!|l 4");
		System.out.println("          3. java SFE.PF.BobPF -s Maximum.txt bQ91:d_aV!|l 4 UCpract");
		System.exit(1);
	}

	//---------------------------------------------------------------
 
	/**
	 * Main program for activating BobPF
	 *
	 * @param args - command line arguments.<br>
	 *               args[0] should be -c, -r<br>
	 *               args[1] should be filename<br>
	 *               args[2] should be seed for RNG<br>
	 *               args[3] should be ot_type (only with -r)<br>
	 *               args[4] can be type of UC
	 *
	 * @see SFE.PF.BobPF#bobUsage
	 */
	public static void main(String[] args) throws Exception {
		String filename;
		String circ_fname;
		String fmt_fname;
		boolean compile = false;
		boolean run_stats = false;
		boolean run = false;

		// Load logging configuration file
		PropertyConfigurator.configure(MyUtil.pathFile("SFE_logcfg.lcf"));

		// Various legality tests on command line parameters
		if ((args.length < 2) || (args.length > 5))
			bobUsage(1);

		compile = args[0].equals("-c");
		run_stats = args[0].equals("-s");
		run = run_stats || args[0].equals("-r");

		if (!compile && !run)
			bobUsage(2);

		if (run && (args.length < 4))
			bobUsage(3);

		filename = new String(args[1]);
		circ_fname = new String(filename + ".shdlPF.Bob.Opt.circuit");
		fmt_fname = new String(filename + ".shdlPF.Bob.Opt.fmt");

		if (compile) {
			File f = new File(filename);

			if (!f.exists()) {
				System.out.println("Input file " + filename + " not found");
				bobUsage(4);
			}

			SFECompiler.compile(filename, true);
			
			SHDLtoSHDLPF s=new SHDLtoSHDLPF();
			s.convert(filename);
		}

		else if (run) {				
			File f1 = new File(circ_fname);
			File f2 = new File(fmt_fname);

			if (!f1.exists() || !f2.exists()) {
				if (!f1.exists())
					System.out.println("Input file " + circ_fname + " not found");
				if (!f2.exists())
					System.out.println("Input file " + fmt_fname + " not found");
				bobUsage(5);
			}

			System.out.println("Running BobPF...");

			String UC_type;
			if(args.length==5)
				UC_type=args[4];
			else
				UC_type=null;

			try {
				new BobPF(circ_fname, fmt_fname, args[2], args[3], run_stats, UC_type);
			} catch (Exception e) {
				System.out.println("Bob's main err: " + e.getMessage());
				e.printStackTrace();
			}		
		}
	}
}
