//SHDLtoSHDLPF.java - convert optimized SHDL to SHDLPF
//Copyright (c) 2007 Thomas Schneider
//See full Copyright license terms in file ../GPL.txt

package SFE.PF;

import java.io.*;
import java.util.Vector;

import org.apache.log4j.*;

import SFE.BOAL.Formatter;
import SFE.BOAL.FormatterError;
import SFE.BOAL.IO;
import SFE.BOAL.MyUtil;
import SFE.BOAL.Parser;

/**
 * Convert optimized SHDL to SHDLPF.
 * 
 * SHDLPF is a subset of SHDL consisting of inputs and 2-input gates (that may be outputs) only.
 *
 * @author Thomas Schneider
 */
class SHDLtoSHDLPF {
	private static final Logger logger = Logger.getLogger(SHDLtoSHDLPF.class);

	int nr_inputs, nr_outputs, nr_gates;
	
	private Vector translation=new Vector(1000,1000);
	
	public int translate(int i) {
		return ((Integer)translation.get(i)).intValue();
	}
	
	private void addtranslation(int from, int to) {
        translation.add(new Integer(to));
    	logger.debug("translate "+from+" -> "+to);
	}

	public class Parser1to2 extends Parser {
		private PrintWriter out=null;
		public int extragates=0;

		class OutputLine {
			int linenr;
			boolean is_output;
			boolean is_input;
			int[] table;
			int[] inputs;

			public OutputLine(int linenr, boolean is_output, int[] table, int[] inputs) {
				this.is_input=false;
				this.is_output=is_output;
				this.linenr=linenr;
				this.table=table;
				this.inputs=inputs;
			}
			
			public OutputLine(int linenr) {
				this.is_input=true;
				this.linenr=linenr;
			}
			
			public void print(PrintWriter out) {
		    	out.print(linenr);
		    	if(is_input) {
		    		out.println(" input");
		    		nr_inputs++;
		    	}
		    	else {
			    	if(is_output) {
			    		out.print(" output");
			    		nr_outputs++;
			    	}
			    	nr_gates++;
			        out.print(" gate arity "+inputs.length+ " table [");
			        for(int i=0; i<table.length; i++)
			        	out.print(" "+table[i]);
			        out.print(" ] inputs [");
			        for(int i=0; i<inputs.length; i++)
			        	out.print(" "+inputs[i]);
			        out.println(" ]");
		    	}
			}
		}
		
		private Vector OutputLines=new Vector(1000,1000);
				
	    private int[] parseInputs() throws ParseError {
	    	int x[]=new int[ninputs];
	    	
	        for (int i = 0; i < ninputs; i++) {
	            x[i]=parseWireNum();
	        }

	        return x;
	    }
		
	    private int[] parseTruthTable() throws ParseError {
	        int out_bit;

	        int x[]=new int[1<<ninputs];
	    	        
	        // expecting 2^ninputs output bits (0/1)
	        for (int i = 0; i < (1 << ninputs); i++) {
	            out_bit = parseToken(false);
	            
	            if ((out_bit != 0) && (out_bit != 1)) {
	                throw new ParseError("ParseOutputs: bad output bit at " +
	                    line_num);
	            }
	            
	            x[i]=out_bit;
	        }

	        return x;
	    }
	    
	    private void outputGate(boolean out_line, int[] tt, int[] in) {
	    	OutputLines.add(new OutputLine(line_num+extragates, out_line, tt, in));
	    }
	    
		// replace gate by multiple 2-input gates using Shannon's expansion theorem
	    private void outputGateN(boolean out_line, int[] tt, int[] in) {
	    	int base=line_num+extragates;
	    	int basedelta=1;
	    	
	    	int[] lsbs={in[0], in[1]};
	    	final int[] tt1={0,0,1,0};
	    	final int[] tt2={0,0,0,1};
	    	final int[] tt3={0,1,1,1};
	    	
	    	// 2-input gates for 2 lsb inputs
	    	for(int k=0; k<(1 << (in.length-2)); k++) {
	    		outputGate(false,new int[]{tt[4*k], tt[4*k+1], tt[4*k+2], tt[4*k+3]}, lsbs);
	    		extragates++;
	    	}
	    	
	    	// MUXes for all other inputs
	    	for(int i=2; i<ninputs; i++) {
	    		for(int k=0; k<(1<<(ninputs-i-1)); k++) {
	        		outputGate(false, tt1, new int[]{in[i],base});
	        		base+=basedelta; extragates++;

	        		outputGate(false, tt2, new int[]{in[i],base});
	        		base+=basedelta; extragates++;
	        		
	        		outputGate((i==ninputs-1&&out_line), tt3, new int[]{line_num+extragates-2,line_num+extragates-1});
	        		extragates++;
	    		}
	    		basedelta=3;
	    		base+=2;
	    	}
	    	extragates--;
	    }
	    
		private void parseGate2(boolean out_line) throws ParseError {
	        int[] truthTable, inputs;
	        
	        parseString("arity");
	        ninputs = parseNumInputs();
	        parseString("table");
	        truthTable=parseTruthTable();
	        parseString("inputs");
	        inputs=parseInputs();
	        
	        switch(ninputs) {
	        case 0:
	    		throw new ParseError("parseGate2 : Please optimize input circuit to contain no 0-input gates at line " +
	                    line_num);
	        	
	        case 1:
	        	// replace 1-input gate with equivalent 2-input gate
	        	int[] tt=new int[4];
	        	tt[0]=truthTable[0];
	        	tt[3]=truthTable[1];
	        	tt[1]=tt[2]=0; // dummy values that will never occur
	        	int in[]={translate(inputs[0]),translate(inputs[0])};
	        	outputGate(out_line,tt,in);

	        	// add gate to translation table
	        	addtranslation(line_num,line_num+extragates);

				break;
	    		
	        case 2:
				for(int i=0; i<inputs.length; i++) inputs[i]=translate(inputs[i]);        	
	        	outputGate(out_line,truthTable,inputs);
	 
	        	// add last gate to translation table
	        	addtranslation(line_num,line_num+extragates);
	        	break;
	        		
	        default:
				for(int i=0; i<inputs.length; i++) inputs[i]=translate(inputs[i]);        	
	        	outputGateN(out_line,truthTable,inputs);
	                	
	        	// add last gate to translation table
	        	addtranslation(line_num,line_num+extragates);        
	        	break;
	        }
	    }
		
	    public void parse() throws ParseError {
	        boolean out_line;
	        int line_type;

	        try {
	            while (parseLineNum()) {
	                out_line = parseOptString("output");
	                line_type = parseLineName();

	                if (line_type == INPUT_LINE) {
	                	OutputLines.add(new OutputLine(line_num+extragates));
	                    // add last gate to translation table
	                	addtranslation(line_num,(line_num+extragates));
	                } else {
	                    parseGate2(out_line);
	                }
	            }
	        } catch (Exception e) {
	            logger.error("parse: exception " + e.getMessage());
	        }
	        logger.debug("Total number of inputs and gates: "+line_num+" -> "+(line_num+extragates));
	    }
	    
	    public void output() {
	    	for(int i=0; i<OutputLines.size(); i++)
	    		((OutputLine)OutputLines.get(i)).print(out);
	    }

		public Parser1to2(StreamTokenizer st, PrintWriter pw) {
			super(st);
			out=pw;
		}
	}	
	
	class Formatter1to2 extends Formatter {
		private static final long serialVersionUID = 3904671386641134632L;
		
		public void parse(PrintWriter outAlice, PrintWriter outBob) throws FormatterError {
			super.parse();
			
		    // Output for Bob scan through all format objects for input objects
		    for (int i = 0; i < FMT.size(); i++) {

		        IO io = (IO) FMT.elementAt(i);

		        if(io.isAlice()) outBob.print("Alice");
		        else outBob.print("Bob");

	        	if(io.isInput()) outBob.print(" input");
		        else outBob.print(" output");
		        
		        outBob.print(" integer");
		        
		        outBob.print(" \""+io.getPrefix()+"\"");
		        
		        outBob.print(" [");
		        for (int j = 0; j < io.getNLines(); j++) {
		        	int num=io.getLinenum(j);
		        	outBob.print(" "+translate(num));
		        }
		        outBob.println(" ]");
		    }

		    // Alice Inputs
		    for (int i = 0; i < FMT.size(); i++) {
		        IO io = (IO) FMT.elementAt(i);
		        if(io.isAlice() && io.isInput()) {
		        	outAlice.println("Alice input integer \""+io.getPrefix()+"\" "+io.getNLines());
		        }
		    }

		    // Alice UC
		    outAlice.println("UC inputs "+nr_inputs+" outputs "+nr_outputs+" gates "+nr_gates);

		    // Alice Outputs
		    for (int i = 0; i < FMT.size(); i++) {
		        IO io = (IO) FMT.elementAt(i);
		        if(io.isAlice() && !io.isInput()) {
		        	outAlice.println("Alice output integer \""+io.getPrefix()+"\" "+io.getNLines());
		        }
		    }
		}
		
		public Formatter1to2(StreamTokenizer st) {
			super(st);
		}
	}
	
	public static void shdlPFusage() {
		System.out.println("Usage: java SFE.PF.SHDLtoSHDLPF <SHDLfilename>");
		System.exit(1);
	}

	public void convert(String filename) {
		Parser1to2 p = null;	// the parser and builder of the circuit
		Formatter1to2 f = null;		// the parser and builder of the IO		
		
		String circuit_filename = filename + ".Opt.circuit";
		String fmt_filename = filename + ".Opt.fmt";

		String circuitBob_filename = filename + ".shdlPF.Bob.Opt.circuit";
		String fmtBob_filename = filename + ".shdlPF.Bob.Opt.fmt";
		String fmtAlice_filename = filename + ".shdlPF.Alice.Opt.fmt";
		
		FileReader fr = null;
		StreamTokenizer st = null;
		PrintWriter pwCircuit=null, pwAlice=null, pwBob=null;

		// --------------------
		// Convert circuit file
		// --------------------
		System.out.println("Converting circuit file to SHDLPF.");
		
		// Preparations
		try {
			fr = new FileReader(circuit_filename);
			st = new StreamTokenizer(fr);
		} catch (IOException e) {
			logger.error("main: cannot open " + circuit_filename + " - " +
					e.getMessage());
		}
		try {
			pwCircuit = new PrintWriter(new FileWriter(circuitBob_filename));
		} catch (IOException e) {
			logger.error("main: cannot open " + circuitBob_filename + " - " +
					e.getMessage());
		}
		
		try {
			// Parse
			p = new Parser1to2(st,pwCircuit);
			p.parse();
			p.output();

			// Clean-up
			fr.close();
			pwCircuit.close();        
		} catch (Exception e) {
			logger.error("main: exception - " + e.getMessage());
			return;
		}
		System.out.println("Completed.");
	
		
		// -------------------
		// Convert format file
		// -------------------
		System.out.println("Converting format file to SHDLPF.");

		// Preparations
		try {
			fr = new FileReader(fmt_filename);
			st = new StreamTokenizer(fr);
		} catch (IOException e) {
			logger.error("main: cannot open " + fmt_filename + " - " +
					e.getMessage());
		}
		try {
			pwAlice = new PrintWriter(new FileWriter(fmtAlice_filename));
		} catch (IOException e) {
			logger.error("main: cannot open " + fmtAlice_filename + " - " +
					e.getMessage());
		}		
		try {
			pwBob = new PrintWriter(new FileWriter(fmtBob_filename));
		} catch (IOException e) {
			logger.error("main: cannot open " + fmtBob_filename + " - " +
					e.getMessage());
		}
		
		try {
			// Parse
			f = new Formatter1to2(st);
			f.parse(pwAlice, pwBob);

			// Clean-up
			fr.close();
			pwAlice.close();
			pwBob.close();
		} catch (Exception e) {
			logger.error("main: exception - " + e.getMessage());
		}
		System.out.println("Completed.");
	}
	
	/**
	 * Convert SHDL files to SHDLPF files
	 *
	 * @param args - command line arguments.
	 *               args[0] should be SHDL filename
	 */
	public static void main(String[] args) {
		// Load logging configuration file
		PropertyConfigurator.configure(MyUtil.pathFile("SFE_logcfg.lcf"));

		if (args.length != 1) 
			shdlPFusage();

		SHDLtoSHDLPF s=new SHDLtoSHDLPF();
		s.convert(args[0]);
	}
}
